import os

import pandas as pd

from Bio.SeqIO.FastaIO import FastaIterator
from Bio.SeqRecord import SeqRecord


class FastaReader:
    """
    Universal class for reading FASTA files with genome or protein
    sequence or multi-FASTA with chunks of sequences

    Example:

        fname = 'NC_001604.fasta'
        fr = FastaReader(fname)

        kmers_sequence = fr.get_sequence()

        ks_df = fr.to_df()
    """

    def __init__(self, fasta_file_path: str):
        self.fasta_file_path = fasta_file_path
        self.fasta_name = os.path.basename(self.fasta_file_path)

    @staticmethod
    def _fasta_reader(filename: str) -> SeqRecord:
        """
        FASTA file reader as iterator
        """

        with open(filename) as handle:
            for record in FastaIterator(handle):
                yield record

    @staticmethod
    def _normalize(entry: SeqRecord) -> str:
        """
        Each of the sequence is normalized into uppercase
        format without blank chars at the end
        """

        return str(entry.seq).upper().strip()

    def get_sequence(self) -> str:
        """
        Final genome or protein sequence string after normalization
        """

        sequence: str = ""

        for entry in self._fasta_reader(self.fasta_file_path):
            sequence += f"{self._normalize(entry)} "

        return sequence.strip()

    def to_df(self) -> pd.DataFrame:
        """
        Return pandas DataFrame with k-mers sequence
        format what is expected by KMersTransformer
        """

        return pd.DataFrame(data=[self.get_sequence()], columns=["sequence"])



import os
import base64
from io import BytesIO
from zipfile import ZipFile
from typing import Dict
from pathlib import Path

import requests

from gensim.models.word2vec import Word2Vec

from fake_useragent import UserAgent


class Word2VecModelManager:
    """
    Manager class is responsible to download and unzip
    Word2Vec pre-trained model for nucleotides embedding
    """

    WORD2VEC_URL = base64.b64decode(
        "aHR0cHM6Ly9kZWVwcGV0cmkuYWkvc3RhdGljL3BoYWdlczIwNTAv"
        "d29yZDJ2ZWMtZW1iZWRkaW5nLTIxLjA3LjIwMjAuemlw"
    )
    STATUS_CODE_200 = 200

    def __init__(self, model_dir: str = "word2vec_model"):
        self.model_dir = model_dir

        if not os.path.exists(model_dir):
            os.mkdir(self.model_dir)

    @staticmethod
    def _get_headers() -> Dict:
        """
        Return header dict with random User-Agent to support request
        and to avoid being blocked by the server
        """

        ua = UserAgent()
        ua.update()

        return {"User-Agent": ua.random}

    def download_model(self) -> Path:
        """
        Download Word2Vec pre-trained model and unzip it into directory

        This procedure should be executed once and the result
        loaded by Word2VecEmbedding class instance
        """

        path = Path(self.model_dir)
        # If model directory exists then return it immediately
        if os.path.exists(path):
            print("[DEBUG] Word2Vec model exists")
            return path
        else:
            print("[DEBUG] Word2Vec model is downloading now")

        headers = self._get_headers()

        with requests.get(self.WORD2VEC_URL, headers=headers) as response:
            assert response.status_code == self.STATUS_CODE_200

            with ZipFile(BytesIO(response.content)) as zip_file:
                zip_file.extractall(self.model_dir)

        return path


class Word2VecEmbedding:
    """
    Word2Vec instance loader class
    """

    def __init__(self, model_pkl_file: str):
        """
        Pickle file need to be serialized by Word2Vec.save method
        before it will be loader with this class
        """

        self.model_pkl_file = model_pkl_file
        if not os.path.exists(self.model_pkl_file):
            raise Exception("Word2Vec model wasn't downloaded yet")

        self.model = Word2Vec.load(self.model_pkl_file)
        self.feature_space = self.model.vector_size

    def get_train_params(self) -> Exception:
        """
        TODO: return dict with model train parameters
        """
        raise NotImplemented



from typing import Set, Union, List

import numpy as np
import pandas as pd

from sklearn.base import BaseEstimator, TransformerMixin
from pandas.core.series import Series

from gensim.models.word2vec import Word2Vec
from gensim.models.fasttext import FastText

from pandarallel import pandarallel


# Parallelization has a cost, so parallelization is efficient only
# if the amount of calculation to parallelize is high enough.
# For very little amount of data, using parallelization is not always worth it.
pandarallel.initialize()


class KMersTransformer(BaseEstimator, TransformerMixin):
    """
    K-mer transformer is responsible to extract set of
    words which are subsequences of length (6 by default)
    contained within a biological sequence

    Each of the word is called k-mer and are composed of
    nucleotides (i.e. A, T, G, and C)

    Example:
        fname = 'NC_001604.fasta'
        fr = FastaReader(fname)

        sample = fr.to_df()

        kmt = KMersTransformer()
        kmt.transform(sample)
    """

    def __init__(self, size: int = 6):
        self.accepted_chars: Set[str] = {"A", "C", "T", "G"}
        self.size: int = size

    def _extract_kmers_from_sequence(self, sequence: str) -> str:
        """
        K-mer transformer with sliding window method,
        where each k-mer has size of 6 (by default)

        A sliding window is used to scan the entire sequence,
        if the k-mer contains unsupported character then the
        whole k-mer is ignored (not included in final string)

        Method return a string with k-mers separated by space
        what is expected as input for embedding
        """

        return " ".join(
            [
                sequence[x : x + self.size]
                for x in range(len(sequence) - self.size + 1)
                if not set(sequence[x : x + self.size]) - self.accepted_chars
            ]
        )

    def transform(self, df: pd.DataFrame) -> Series:
        """
        Execute k-mer transformer on each DNA sequence
        and return it as Series with k-mers strings
        """

        # sequence column is expected
        assert list(df.columns) == ["sequence"]

        return df.sequence.parallel_apply(self._extract_kmers_from_sequence)


class GenomeAvgTransformer(TransformerMixin, BaseEstimator):
    """
    Average k-mers to represent Bacteriophage with word embedding

    Most Word2vec or fastText pre-trained models allow to get
    numerical representations of individual words but not of entire documents
    With this class it can average each k-mer of a DNA so that the
    generated Bacteriophage vector is actually a centroid of all k-mers in feature space
    """

    def __init__(self, gensim_model: Union[FastText, Word2Vec]):
        """
        It support Word2Vec as well as fastText embedding model
        """

        self.gensim_model: Union[FastText, Word2Vec] = gensim_model
        self.columns = [
            f"feature_{index}" for index in range(self.gensim_model.vector_size)
        ]

    def average_word_vectors(self, words: List[str], vocabulary: Set) -> np.array:
        """
        Return fixed-length numeric vector for each DNA sequence
        """

        # Filter only words supported by the vocabulary
        supported_words: list = [word for word in words if word in vocabulary]

        if supported_words:
            # Return average fixed-length numeric vector
            # including all the k-mers in the sequence
            feature_vector: np.array = np.mean(
                self.gensim_model[supported_words], axis=0, dtype="float64"
            )
        else:
            # Return fixed-length numeric vector with zeros
            # if the k-mers (words) weren't in the vocabulary
            feature_vector: np.array = np.zeros(
                (self.gensim_model.vector_size,), dtype="float64"
            )

        return feature_vector

    def averaged_word_vectorizer(self, column_with_kmers_seqs) -> np.array:
        """
        Execute DNA averaged vector transformer on each k-mer sequence
        and return as array of numeric values
        """

        # Unique set of words
        vocabulary: set = set(self.gensim_model.wv.index2word)

        features: list = [
            self.average_word_vectors(
                # Split k-mer sequence with spaces into a list with k-mers (words)
                words=sentence.split(),
                vocabulary=vocabulary,
            )
            for sentence in column_with_kmers_seqs
        ]

        return np.array(features)

    def transform(self, column_with_kmers_seqs: Series) -> pd.DataFrame:
        """
        Execute DNA averaged vector transformer on each k-mer sequence
        and return it Pandas DataFrame with fixed-length numeric vector space
        """

        return pd.DataFrame(
            data=self.averaged_word_vectorizer(column_with_kmers_seqs),
            columns=self.columns,
        )
